#include <sys/reboot.h>
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/epoll.h>

typedef struct th_data_t
{
    int id;
    int epfd;
} th_data_t;

#define _EXIT                   \
    do {                        \
        reboot(RB_POWER_OFF);   \
        return 0;               \
    } while(0)

void *epoll_wait_thread(void *param)
{
    th_data_t *th_data = param;
    struct epoll_event ev = {0};
    int rc = 0;

    printf("thread %d: Start polling\n", th_data->id);
    rc = epoll_wait(th_data->epfd, &ev, 1, 3000);
    printf("thread %d: Got %d events\n", th_data->id, rc);

    return NULL;
}

int main(int argc, char **argv)
{
    pthread_t th1, th2;
    int s[2];
    int data = 0xdeadbeef;
    int epfd = epoll_create1(0);
    struct epoll_event ev;
    th_data_t th_data[2];

    if (epfd == -1)
    {
        perror("epoll_create1");
        _EXIT;
    }

    /* Create a connection. */
    if (socketpair(AF_UNIX, SOCK_DGRAM, 0, s) != 0)
    {
        perror("socketpair");
        _EXIT;
    }

    /* Create epoll descriptor and register a socket. */
    ev.events = EPOLLIN;
    ev.data.fd = s[0];
    if (epoll_ctl(epfd, EPOLL_CTL_ADD, s[0], &ev) != 0)
    {
        perror("epoll_ctl");
        _EXIT;
    }

    /* Start 2 threads with epoll_wait() routine. */
    th_data[0].id = 0;
    th_data[0].epfd = epfd;
    pthread_create(&th1, NULL, epoll_wait_thread, &th_data[0]);

    th_data[1].id = 1;
    th_data[1].epfd = epfd;
    pthread_create(&th2, NULL, epoll_wait_thread, &th_data[1]);

    sleep(1);

    /* Send some data to unblock epoll_wait() threads. */
    puts("main: Send some data");
    if (write(s[1], &data, sizeof(data)) != sizeof(data))
    {
        perror("write");
        _EXIT;
    }

    /* Cleanup. */
    pthread_join(th1, NULL);
    pthread_join(th2, NULL);

    _EXIT;
}

